"""
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

import os
import sys
import re
import configparser as ConfigParser

class parse_config(dict):
    cp = ConfigParser.ConfigParser()
    def __init__(self,configfile='proxy.ini'):
        if os.access(configfile, os.R_OK) == False:
            print("Unable to read config file: %s" % configfile)
            sys.exit(2)
        try:
            self.cp.read(configfile)
        except ConfigParser.Error:
            print('Unable to parse config file: %s' % configfile)
            sys.exit(2)
        for section in self.cp.sections():
            if section in ['proxy', 'pxe', 'ipxe']:
                self[section]={}
                for item in self.cp.items(section):
                    value = self.cp.get(section,item[0])
                    valuecheckmsg = 'Please check the value set for ' + item[0] + ' in the ' + section + ' section of ' + configfile
                    if item[0] == 'listen_address':
                        if self.listenAddressCheck(value):
                            self[section][item[0]] = value
                        else:
                            print(valuecheckmsg)
                            sys.exit(2)
                    elif item[0] == 'tftpd':
                        if self.ipAddressCheck(value):
                            self[section][item[0]] = list(map(int, value.split(".")))
                        else:
                            print(valuecheckmsg)
                            sys.exit(2)
                    elif item[0] in ['filename', 'filename_efi32', 'filename_efi64', 'interface', 'legacy_bootstrap', 'efi_bootstrap', 'chainload_url']:
                        if self.stringCheck(value):
                            self[section][item[0]] = value
                        else:
                            print(valuecheckmsg)
                            sys.exit(2)
                    elif item[0] in ['enabled']:
                        self[section][item[0]] = value
                    elif item[0] in ['vendor_specific_information']:
                        if self.stringCheck(value):
                            self[section][item[0]] = value
                        else:
                            print(valuecheckmsg)
                            sys.exit(2)
                    else:
                            print('The item ' + item[0] + ' in the ' + section + ' section of ' + configfile + ' is unknown')
                            sys.exit(2)
            else:
                print('The ' + section + ' section in ' + configfile + ' is unknown')
                exit(2)
            self['proxy']['client_listen_port'] = "68"
            self['proxy']['server_listen_port'] = "67"
            
        # Single Source of Truth / Converge config IPs
        from . import net
        conf_ip = self['proxy'].get("listen_address")
        conf_iface = self['proxy'].get("interface")
        
        if conf_iface and conf_ip:
            if_ip = net.get_ip_address(conf_iface)
            if if_ip != conf_ip:
                print("CRITICAL: Split-Brain Config! listen_address (%s) does not match IP of interface %s (%s)." % (conf_ip, conf_iface, if_ip))
                sys.exit(1)
        elif conf_iface:
            if_ip = net.get_ip_address(conf_iface)
            if not if_ip:
                print("CRITICAL: Cannot resolve IP for interface %s" % conf_iface)
                sys.exit(1)
            self['proxy']["listen_address"] = if_ip
        elif conf_ip:
            if_name = net.get_dev_name(conf_ip)
            if not if_name and conf_ip != "0.0.0.0":
                print("CRITICAL: Cannot resolve interface for IP %s" % conf_ip)
                sys.exit(1)
            self['proxy']["interface"] = if_name
        else:
            print("CRITICAL: You must provide either 'interface' or 'listen_address' in proxy.ini.")
            sys.exit(1)
                
    def ipAddressCheck(self,ip_str):
        pattern = r"\b(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\b"
        if re.match(pattern, ip_str):
            return True
        else:
            return False
    
    def listenAddressCheck(self,ip_str):
        if (self.ipAddressCheck(ip_str) or "0.0.0.0" == ip_str):
            return True
        else:
            return False
    
    def intCheck(self,input):
        try:
            int(input)
        except ValueError:
            return False
        return True
    
    def stringCheck(self,input):
        if (type(input) == str and len(input) > 0):
            return True
        else:
            return False
            
if __name__ == "__main__":
    configp=parse_config()
    print(configp)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2005, TUBITAK/UEKAE
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version. Please read the COPYING file.
#

import os
import array
import fcntl
import struct
import socket

class interface:
    """ ioctl stuff """

    IFNAMSIZ = 16               # interface name size

    # From <bits/ioctls.h>

    SIOCGIFADDR = 0x8915        # get PA address
    SIOCGIFBRDADDR  = 0x8919    # get broadcast PA address
    SIOCGIFCONF = 0x8912        # get iface list
    SIOCGIFFLAGS = 0x8913       # get flags
    SIOCGIFMTU = 0x8921         # get MTU size
    SIOCGIFNETMASK  = 0x891b    # get network PA mask
    SIOCSIFADDR = 0x8916        # set PA address
    SIOCSIFBRDADDR  = 0x891a    # set broadcast PA address
    SIOCSIFFLAGS = 0x8914       # set flags
    SIOCSIFMTU = 0x8922         # set MTU size
    SIOCSIFNETMASK  = 0x891c    # set network PA mask

    # From <net/if.h>    

    IFF_UP = 0x1           # Interface is up.
    IFF_BROADCAST = 0x2    # Broadcast address valid.
    IFF_DEBUG = 0x4        # Turn on debugging.
    IFF_LOOPBACK = 0x8     # Is a loopback net.
    IFF_POINTOPOINT = 0x10 # Interface is point-to-point link.
    IFF_NOTRAILERS = 0x20  # Avoid use of trailers.
    IFF_RUNNING = 0x40     # Resources allocated.
    IFF_NOARP = 0x80       # No address resolution protocol.
    IFF_PROMISC = 0x100    # Receive all packets.
    IFF_ALLMULTI = 0x200   # Receive all multicast packets.
    IFF_MASTER = 0x400     # Master of a load balancer.
    IFF_SLAVE = 0x800      # Slave of a load balancer.
    IFF_MULTICAST = 0x1000 # Supports multicast.
    IFF_PORTSEL = 0x2000   # Can set media type.
    IFF_AUTOMEDIA = 0x4000 # Auto media select active.


    def __init__(self):
        # create a socket to communicate with system
        self.sockfd = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    def _ioctl(self, func, args):
        return fcntl.ioctl(self.sockfd.fileno(), func, args)

    def _call(self, ifname, func, ip = None):

        if isinstance(ifname, str):
            ifreq = ifname.encode('utf-8')
        else:
            ifreq = ifname

        if ip is None:
            data = (ifreq + b'\0'*32)[:32]
        else:
            ifreq = (ifreq + b'\0' * self.IFNAMSIZ)[:self.IFNAMSIZ]
            data = struct.pack("16si4s10x", ifreq, socket.AF_INET, socket.inet_aton(ip))

        try:
            result = self._ioctl(func, data)
        except IOError:
            return None

        return result

    def getInterfaceList(self):
        """ Get all interface names in a list """
        # get interface list
        buffer = array.array('B', b'\0' * 1024)
        ifconf = struct.pack("iP", buffer.buffer_info()[1], buffer.buffer_info()[0])
        result = self._ioctl(self.SIOCGIFCONF, ifconf)

        # loop over interface names
        iflist = []
        size, ptr = struct.unpack("iP", result)
        for idx in range(0, size, 32):
            ifconf = buffer.tobytes()[idx:idx+32]
            name, dummy = struct.unpack("16s16s", ifconf)
            name, dummy = name.split(b'\0', 1)
            iflist.append(name.decode('utf-8'))

        return iflist

    def getAddr(self, ifname):
        """ Get the inet addr for an interface """
        result = self._call(ifname, self.SIOCGIFADDR)
        return socket.inet_ntoa(result[20:24])

    def getNetmask(self, ifname):
        """ Get the netmask for an interface """
        result = self._call(ifname, self.SIOCGIFNETMASK)
        return socket.inet_ntoa(result[20:24])

    def getBroadcast(self, ifname):
        """ Get the broadcast addr for an interface """
        result = self._call(ifname, self.SIOCGIFBRDADDR)
        return socket.inet_ntoa(result[20:24])

    def getStatus(self, ifname):
        """ Check whether interface is UP """
        result = self._call(ifname, self.SIOCGIFFLAGS)
        flags, = struct.unpack('H', result[16:18])
        return (flags & self.IFF_UP) != 0

    def getMTU(self, ifname):
        """ Get the MTU size of an interface """
        data = self._call(ifname, self.SIOCGIFMTU)
        mtu = struct.unpack("16si12x", data)[1]
        return mtu

    def setAddr(self, ifname, ip):
        """ Set the inet addr for an interface """
        result = self._call(ifname, self.SIOCSIFADDR, ip)

        if result and socket.inet_ntoa(result[20:24]) is ip:
            return True
        else:
            return None

    def setNetmask(self, ifname, ip):
        """ Set the netmask for an interface """
        result = self._call(ifname, self.SIOCSIFNETMASK, ip)

        if result and socket.inet_ntoa(result[20:24]) is ip:
            return True
        else:
            return None

    def setBroadcast(self, ifname, ip):
        """ Set the broadcast addr for an interface """
        result = self._call(ifname, self.SIOCSIFBRDADDR, ip)

        if socket.inet_ntoa(result[20:24]) is ip:
            return True
        else:
            return None

    def setStatusDown(self, ifname):
        """ Set interface status (UP/DOWN) """
        if isinstance(ifname, str):
            ifreq = ifname.encode('utf-8')
        else:
            ifreq = ifname
        ifreq = (ifreq + b'\0' * self.IFNAMSIZ)[:self.IFNAMSIZ]

        result = self._call(ifname, self.SIOCGIFFLAGS)
        flags, = struct.unpack('H', result[16:18])
        flags &= ~self.IFF_UP

        data = struct.pack("16sh", ifreq, flags)
        result = self._ioctl(self.SIOCSIFFLAGS, data)

        return result

    def setStatusUp(self, ifname):
        """ Set interface status (UP/DOWN) """
        if isinstance(ifname, str):
            ifreq = ifname.encode('utf-8')
        else:
            ifreq = ifname
        ifreq = (ifreq + b'\0' * self.IFNAMSIZ)[:self.IFNAMSIZ]

        flags = self.IFF_UP
        flags |= self.IFF_RUNNING
        flags |= self.IFF_BROADCAST
        flags |= self.IFF_MULTICAST
        flags &= ~self.IFF_NOARP
        flags &= ~self.IFF_PROMISC

        data = struct.pack("16sh", ifreq, flags)
        result = self._ioctl(self.SIOCSIFFLAGS, data)

        return result

    def setMTU(self, ifname, mtu):
        """ Set the MTU size of an interface """
        if isinstance(ifname, str):
            ifreq = ifname.encode('utf-8')
        else:
            ifreq = ifname
        ifreq = (ifreq + b'\0' * self.IFNAMSIZ)[:self.IFNAMSIZ]

        data = struct.pack("16si", ifreq, mtu)
        result = self._ioctl(self.SIOCSIFMTU, data)

        if struct.unpack("16si", result)[1] is mtu:
            return True
        else:
            return None




# pydhcplib
# Copyright (C) 2008 Mathieu Ignacio -- mignacio@april.org
#
# This file is part of pydhcplib.
# Pydhcplib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import operator
from struct import unpack
from struct import pack
from .dhcp_constants import *
import sys


# DhcpPacket : base class to encode/decode dhcp packets.

class DhcpBasicPacket:
    def __init__(self):
        self.packet_data = [0]*240
        self.options_data = {}
        self.packet_data[236:240] = MagicCookie
        self.source_address = False
        
    def IsDhcpPacket(self):
        if self.packet_data[236:240] != MagicCookie : return False
        return True

    # Check if variable is a list with int between 0 and 255
    def CheckType(self,variable):
        if type(variable) == list :
            for each in variable :
                if (type(each) != int)  or (each < 0) or (each > 255) :
                    return False
            return True
        else : return False
        

    def DeleteOption(self,name):
        # if name is a standard dhcp field
        # Set field to 0
        if name in DhcpFields :
            begin = DhcpFields[name][0]
            end = DhcpFields[name][0]+DhcpFields[name][1]
            self.packet_data[begin:end] = [0]*DhcpFields[name][1]
            return True

        # if name is a dhcp option
        # delete option from self.option_data
        elif name in self.options_data :
            # forget how to remove a key... try delete
            self.options_data.__delitem__(name)
            return True

        return False

    def GetOption(self,name):
        if name in DhcpFields :
            option_info = DhcpFields[name]
            return self.packet_data[option_info[0]:option_info[0]+option_info[1]]

        elif name in self.options_data :
            return self.options_data[name]

        return []
        

    def SetOption(self,name,value):

        # Basic value checking :
        # has value list a correct length
        
        # if name is a standard dhcp field
        if name in DhcpFields :
            if len(value) != DhcpFields[name][1] :
                sys.stderr.write( "pydhcplib.dhcp_basic_packet.setoption error, bad option length : "+name)
                return False
            begin = DhcpFields[name][0]
            end = DhcpFields[name][0]+DhcpFields[name][1]
            self.packet_data[begin:end] = value
            return True

        # if name is a dhcp option
        elif name in DhcpOptions :

            # fields_specs : {'option_code':fixed_length,minimum_length,multiple}
            # if fixed_length == 0 : minimum_length and multiple apply
            # else : forget minimum_length and multiple 
            # multiple : length MUST be a multiple of 'multiple'
            # FIXME : this definition should'nt be in dhcp_constants ?
            fields_specs = { "ipv4":[4,0,1], "ipv4+":[0,4,4],
                             "string":[0,0,1], "bool":[1,0,1],
                             "char":[1,0,1], "16-bits":[2,0,1],
                             "32-bits":[4,0,1], "identifier":[0,2,1],
                             "RFC3397":[0,4,1],"none":[0,0,1],"char+":[0,1,1]
                             }
            
            specs = fields_specs[DhcpOptionsTypes[DhcpOptions[name]]]
            length = len(value)
            if (specs[0]!=0 and specs==length) or (specs[1]<=length and length%specs[2]==0):
                self.options_data[name] = value
                return True
            else :
                return False

        sys.stderr.write( "pydhcplib.dhcp_basic_packet.setoption error : unknown option "+name)
        return False



    def IsOption(self,name):
        if name in self.options_data : return True
        elif name in DhcpFields : return True
        else : return False

    # Encode Packet and return it
    def EncodePacket(self):

        # MUST set options in order to respect the RFC (see router option)
        order = {}

        for each in self.options_data.keys() :
            order[DhcpOptions[each]] = []
            order[DhcpOptions[each]].append(DhcpOptions[each])
            order[DhcpOptions[each]].append(len(self.options_data[each]))
            order[DhcpOptions[each]] += self.options_data[each]
            
        options = []

        for each in sorted(order.keys()) : options += (order[each])

        packet = self.packet_data[:240] + options
        packet.append(255) # add end option
        pack_fmt = str(len(packet))+"B"
        
        return pack(pack_fmt,*packet)


    # Insert packet in the object
    def DecodePacket(self,data,debug=False):
        self.packet_data = []
        self.options_data = {}

        if (not data) : return False
        # we transform all data to int list
        unpack_fmt = str(len(data)) + "B"
        for i in unpack(unpack_fmt,data):
            self.packet_data.append(i)

        # Some servers or clients don't place magic cookie immediately
        # after headers and begin options fields only after magic.
        # These 4 lines search magic cookie and begin iterator after.
        iterator = 236
        end_iterator = len(self.packet_data)
        while ( self.packet_data[iterator:iterator+4] != MagicCookie and iterator < end_iterator) :
            iterator += 1
        iterator += 4
        
        # parse extended options

        while iterator < end_iterator :
            if self.packet_data[iterator] == 0 : # pad option
                opt_first = iterator+1
                iterator += 1

            elif self.packet_data[iterator] == 255 :
                self.packet_data = self.packet_data[:240] # base packet length without magic cookie
                return
                
            elif self.packet_data[iterator] in DhcpOptionsTypes and self.packet_data[iterator]!= 255:
                if iterator + 1 >= end_iterator:
                    break # Protect against out-of-bounds read if option length is missing
                opt_len = self.packet_data[iterator+1]
                opt_first = iterator+1
                self.options_data[DhcpOptionsList[self.packet_data[iterator]]] = self.packet_data[opt_first+1:opt_len+opt_first+1]
                iterator += self.packet_data[opt_first] + 2
            else :
                if iterator + 1 >= end_iterator:
                    break # Protect against out-of-bounds read if option length is missing
                opt_first = iterator+1
                iterator += self.packet_data[opt_first] + 2

        # cut packet_data to remove options
        
        self.packet_data = self.packet_data[:240] # base packet length with magic cookie



# pydhcplib
# Copyright (C) 2008 Mathieu Ignacio -- mignacio@april.org
#
# This file is part of pydhcplib.
# Pydhcplib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


from binascii import unhexlify,hexlify

# Check and convert hardware/nic/mac address type
class hwmac:
    def __init__(self,value="00:00:00:00:00:00") :
        self._hw_numlist = []
        self._hw_string = ""
        hw_type = type(value)
        if hw_type == str :
            value = value.strip()
            self._hw_string = value
            self._StringToNumlist(value)
            self._CheckNumList()
        elif hw_type == list :
            self._hw_numlist = value
            self._CheckNumList()
            self._NumlistToString()
        else : raise TypeError('hwmac init : Valid types are str and list')



    # Check if _hw_numlist is valid and raise error if not.
    def _CheckNumList(self) :
        if len(self._hw_numlist) != 6 : raise ValueError("hwmac : wrong list length.")
        for part in self._hw_numlist :
            if type (part) != int : raise TypeError("hwmac : each element of list must be int")
            if part < 0 or part > 255 : raise ValueError("hwmac : need numbers between 0 and 255.")
        return True


    def _StringToNumlist(self,value):
        self._hw_string = self._hw_string.replace("-",":").replace(".",":")
        self._hw_string = self._hw_string.lower()

        for twochar in self._hw_string.split(":"):
            self._hw_numlist.append(ord(unhexlify(twochar)))
            
    # Convert NumList type ip to String type ip
    def _NumlistToString(self) :
        self._hw_string = ":".join([hexlify(bytes([x])).decode('utf-8') for x in self._hw_numlist])

    # Convert String type ip to NumList type ip
    # return ip string
    def str(self) :
        return self._hw_string

    # return ip list (useful for DhcpPacket class)
    def list(self) :
        return self._hw_numlist

    def __hash__(self) :
        return self._hw_string.__hash__()

    def __repr__(self) :
        return self._hw_string

    def __eq__(self,other) :
        return self._hw_string == other

    def __bool__(self) :
        if self._hw_string != "00:00:00:00:00:00" : return True
        return False







# pydhcplib
# Copyright (C) 2008 Mathieu Ignacio -- mignacio@april.org
#
# This file is part of pydhcplib.
# Pydhcplib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import operator
from struct import unpack
from struct import pack
from .dhcp_basic_packet import *
from .dhcp_constants import *
from .type_ipv4 import ipv4
from .type_strlist import strlist
from .type_hwmac import hwmac
import sys

class DhcpPacket(DhcpBasicPacket):
    def str(self):
        # Process headers : 
        printable_data = "# Header fields\n"

        op = self.packet_data[DhcpFields['op'][0]:DhcpFields['op'][0]+DhcpFields['op'][1]]
        printable_data += "op : " + DhcpFieldsName['op'][str(op[0])] + "\n"

        
        for opt in  ['htype','hlen','hops','xid','secs','flags',
                     'ciaddr','yiaddr','siaddr','giaddr','chaddr','sname','file'] :
            begin = DhcpFields[opt][0]
            end = DhcpFields[opt][0]+DhcpFields[opt][1]
            data = self.packet_data[begin:end]
            result = ''
            if DhcpFieldsTypes[opt] == "int" : result = str(data[0])
            elif DhcpFieldsTypes[opt] == "int2" : result = str(data[0]*256+data[1])
            elif DhcpFieldsTypes[opt] == "int4" : result = str(ipv4(data).int())
            elif DhcpFieldsTypes[opt] == "str" :
                for each in data :
                    if each != 0 : result += chr(each)
                    else : break

            elif DhcpFieldsTypes[opt] == "ipv4" : result = ipv4(data).str()
            elif DhcpFieldsTypes[opt] == "hwmac" :
                result = []
                hexsym = ['0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f']
                for iterator in range(6) :
                    result += [str(hexsym[data[iterator]/16]+hexsym[data[iterator]%16])]

                result = ':'.join(result)

            printable_data += opt+" : "+result  + "\n"

        # Process options : 
        printable_data += "# Options fields\n"

        for opt in self.options_data.keys():
            data = self.options_data[opt]
            result = ""
            optnum  = DhcpOptions[opt]
            if opt=='dhcp_message_type' : result = DhcpFieldsName['dhcp_message_type'][str(data[0])]
            elif DhcpOptionsTypes[optnum] == "char" : result = str(data[0])
            elif DhcpOptionsTypes[optnum] == "16-bits" : result = str(data[0]*256+data[0])
            elif DhcpOptionsTypes[optnum] == "32-bits" : result = str(ipv4(data).int())
            elif DhcpOptionsTypes[optnum] == "string" :
                for each in data :
                    if each != 0 : result += chr(each)
                    else : break
        
            elif DhcpOptionsTypes[optnum] == "ipv4" : result = ipv4(data).str()
            elif DhcpOptionsTypes[optnum] == "ipv4+" :
                for i in range(0,len(data),4) :
                    if len(data[i:i+4]) == 4 :
                        result += ipv4(data[i:i+4]).str() + " - "
            elif DhcpOptionsTypes[optnum] == "char+" :
                if optnum == 55 : # parameter_request_list
                    result = ','.join([DhcpOptionsList[each] for each in data])
                else : result += str(data)
                
            printable_data += opt + " : " + result + "\n"

        return printable_data

    def AddLine(self,_string) :
        (parameter,junk,value) = _string.partition(':')
        parameter = parameter.strip()
        # If value begin with a whitespace, remove it, leave others
        if len(value)>0 and value[0] == ' ' : value = value[1:]
        value = self._OptionsToBinary(parameter,value)
        if value : self.SetOption(parameter,value)

    def _OptionsToBinary(self,parameter,value) :
        # Transform textual data into dhcp binary data

        p = parameter.strip()
        # 1- Search for header informations or specific parameter
        if p == 'op' or p == 'htype' :
            value = value.strip()
            if value.isdigit() : return [int(value)]
            try :
                value = DhcpNames[value.strip()]
                return [value]
            except KeyError :
                return [0]

        elif p == 'hlen' or p == 'hops' :
            try :
                value = int(value)
                return [value]
            except ValueError :
                return [0]

        elif p == 'secs' or p == 'flags' :
            try :
                value = ipv4(int(value)).list()
            except ValueError :
                value = [0,0,0,0]

            return value[2:]

        elif p == 'xid' :
            try :
                value = ipv4(int(value)).list()
            except ValueError :
                value = [0,0,0,0]
            return value

        elif p == 'ciaddr' or p == 'yiaddr' or p == 'siaddr' or p == 'giaddr' :
            try :
                ip = ipv4(value).list()
            except ValueError :
                ip = [0,0,0,0]
            return ip
        
        elif p == 'chaddr' :
            try:
                value = hwmac(value).list()+[0]*10
            except (ValueError, TypeError):
                value = [0]*16
            return value
            
        elif p == 'sname' :
            return
        elif p == 'file' :
            return
        elif p == 'parameter_request_list' :
            value = value.strip().split(',')
            tmp = []
            for each in value:
                if each in DhcpOptions : tmp.append(DhcpOptions[each])
            return tmp
        elif  p=='dhcp_message_type' :
            try :
                return [DhcpNames[value]]
            except KeyError:
                return

        # 2- Search for options
        try : option_type = DhcpOptionsTypes[DhcpOptions[parameter]]
        except KeyError : return False

        if option_type == "ipv4" :
            # this is a single ip address
            try :
                binary_value = list(map(int,value.split(".")))
            except ValueError : return False
            
        elif option_type == "ipv4+" :
            # this is multiple ip address
            iplist = value.split(",")
            opt = []
            for single in iplist :
                opt += (ipv4(single).list())
            binary_value = opt

        elif option_type == "32-bits" :
            # This is probably a number...
            try :
                digit = int(value)
                binary_value = [digit>>24&0xFF,(digit>>16)&0xFF,(digit>>8)&0xFF,digit&0xFF]
            except ValueError :
                return False

        elif option_type == "16-bits" :
            try :
                digit = int(value)
                binary_value = [(digit>>8)&0xFF,digit&0xFF]
            except ValueError : return False


        elif option_type == "char" :
            try :
                digit = int(value)
                binary_value = [digit&0xFF]
            except ValueError : return False

        elif option_type == "bool" :
            if value=="False" or value=="false" or value==0 :
                binary_value = [0]
            else : binary_value = [1]
            
        elif option_type == "string" :
            binary_value = strlist(value).list()

        else :
            binary_value = strlist(value).list()
        
        return binary_value
    
    # FIXME: This is called from IsDhcpSomethingPacket, but is this really
    # needed?  Or maybe this testing should be done in
    # DhcpBasicPacket.DecodePacket().

    # Test Packet Type
    def IsDhcpSomethingPacket(self,type):
        if self.IsDhcpPacket() == False : return False
        if self.IsOption("dhcp_message_type") == False : return False
        if self.GetOption("dhcp_message_type") != type : return False
        return True
    
    def IsDhcpDiscoverPacket(self):
        return self.IsDhcpSomethingPacket([1])

    def IsDhcpOfferPacket(self):
        return self.IsDhcpSomethingPacket([2])

    def IsDhcpRequestPacket(self):
        return self.IsDhcpSomethingPacket([3])

    def IsDhcpDeclinePacket(self):
        return self.IsDhcpSomethingPacket([4])

    def IsDhcpAckPacket(self):
        return self.IsDhcpSomethingPacket([5])

    def IsDhcpNackPacket(self):
        return self.IsDhcpSomethingPacket([6])

    def IsDhcpReleasePacket(self):
        return self.IsDhcpSomethingPacket([7])

    def IsDhcpInformPacket(self):
        return self.IsDhcpSomethingPacket([8])


    def GetMultipleOptions(self,options=()):
        result = {}
        for each in options:
            result[each] = self.GetOption(each)
        return result

    def SetMultipleOptions(self,options={}):
        for each in options.keys():
            self.SetOption(each,options[each])






    # Creating Response Packet

    # Server-side functions
    # From RFC 2132 page 28/29
    def CreateDhcpOfferPacketFrom(self,src): # src = discover packet
        self.SetOption("htype",src.GetOption("htype"))
        self.SetOption("xid",src.GetOption("xid"))
        self.SetOption("flags",src.GetOption("flags"))
        self.SetOption("giaddr",src.GetOption("giaddr"))
        self.SetOption("chaddr",src.GetOption("chaddr"))
        self.SetOption("ip_address_lease_time",src.GetOption("ip_address_lease_time"))
        self.TransformToDhcpOfferPacket()

    def TransformToDhcpOfferPacket(self):
        self.SetOption("dhcp_message_type",[2])
        self.SetOption("op",[2])
        self.SetOption("hlen",[6]) 

        self.DeleteOption("secs")
        self.DeleteOption("ciaddr")
        self.DeleteOption("request_ip_address")
        self.DeleteOption("parameter_request_list")
        self.DeleteOption("client_identifier")
        self.DeleteOption("maximum_message_size")





    """ Dhcp ACK packet creation """
    def CreateDhcpAckPacketFrom(self,src): # src = request or inform packet
        self.SetOption("htype",src.GetOption("htype"))
        self.SetOption("xid",src.GetOption("xid"))
        self.SetOption("ciaddr",src.GetOption("ciaddr"))
        self.SetOption("flags",src.GetOption("flags"))
        self.SetOption("giaddr",src.GetOption("giaddr"))
        self.SetOption("chaddr",src.GetOption("chaddr"))
        self.SetOption("ip_address_lease_time",src.GetOption("ip_address_lease_time"))
        self.TransformToDhcpAckPacket()

    def TransformToDhcpAckPacket(self): # src = request or inform packet
        self.SetOption("op",[2])
        self.SetOption("hlen",[6]) 
        self.SetOption("dhcp_message_type",[5])

        self.DeleteOption("secs")
        self.DeleteOption("request_ip_address")
        self.DeleteOption("parameter_request_list")
        self.DeleteOption("client_identifier")
        self.DeleteOption("maximum_message_size")


    """ Dhcp NACK packet creation """
    def CreateDhcpNackPacketFrom(self,src): # src = request or inform packet
        
        self.SetOption("htype",src.GetOption("htype"))
        self.SetOption("xid",src.GetOption("xid"))
        self.SetOption("flags",src.GetOption("flags"))
        self.SetOption("giaddr",src.GetOption("giaddr"))
        self.SetOption("chaddr",src.GetOption("chaddr"))
        self.TransformToDhcpNackPacket()

    def TransformToDhcpNackPacket(self):
        self.SetOption("op",[2])
        self.SetOption("hlen",[6]) 
        self.DeleteOption("secs")
        self.DeleteOption("ciaddr")
        self.DeleteOption("yiaddr")
        self.DeleteOption("siaddr")
        self.DeleteOption("sname")
        self.DeleteOption("file")
        self.DeleteOption("request_ip_address")
        self.DeleteOption("ip_address_lease_time")
        self.DeleteOption("parameter_request_list")
        self.DeleteOption("client_identifier")
        self.DeleteOption("maximum_message_size")
        self.SetOption("dhcp_message_type",[6])







    """ GetClientIdentifier """

    def GetClientIdentifier(self) :
        if self.IsOption("client_identifier") :
            return self.GetOption("client_identifier")
        return []

    def GetGiaddr(self) :
        return self.GetOption("giaddr")

    def GetHardwareAddress(self) :
        length = self.GetOption("hlen")[0]
        full_hw = self.GetOption("chaddr")
        if length!=[] and length<len(full_hw) : return full_hw[0:length]
        return full_hw




import socket
import fcntl
import struct
import array

def get_ip_address(ifname):
    if isinstance(ifname, str):
        ifname = ifname.encode('utf-8')
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        return socket.inet_ntoa(fcntl.ioctl(
            s.fileno(),
            0x8915,  # SIOCGIFADDR
            struct.pack('256s', ifname[:15])
        )[20:24])
    except OSError:
        return None

def all_interfaces():
    return [i[1] for i in socket.if_nameindex()]

def get_dev_name(ipaddr):
    for netdev in all_interfaces():
        if get_ip_address(netdev) == ipaddr:
            return netdev


# pydhcplib
# Copyright (C) 2008 Mathieu Ignacio -- mignacio@april.org
#
# This file is part of pydhcplib.
# Pydhcplib is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

class strlist :
    def __init__(self,data="") :
        str_type = type(data)
        self._str = ""
        self._list = []
        
        if str_type == str :
            self._str = data
            for each in range(len(self._str)) :
                self._list.append(ord(self._str[each]))
        elif str_type == bytes :
            self._str = data.decode("utf-8", "ignore")
            for each in range(len(data)) :
                self._list.append(data[each])
        elif str_type == list :
            self._list = data
            self._str = "".join(map(chr,self._list))
        else : raise TypeError('strlist init : Valid types are str, bytes, and list of int')

    # return string
    def str(self) :
        return self._str

    # return list (useful for DhcpPacket class)
    def list(self) :
        return self._list

    # return int
    # FIXME
    def int(self) :
        return 0



    """ Useful function for native python operations """

    def __hash__(self) :
        return self._str.__hash__()

    def __repr__(self) :
        return self._str

    def __bool__(self) :
        if self._str != "" : return True
        return False

    def __eq__(self,other) :
        return self._str == other
		    





"""

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from .proxyconfig import parse_config
from .dhcplib.dhcp_network import *
from .dhcplib.dhcp_packet import *
import logging
import logging.handlers
import sys
from . import net
import traceback

class DhcpServerBase(DhcpNetwork) :
    def __init__(self, listen_address="0.0.0.0", client_listen_port=68,server_listen_port=67, interface="") :
        
        DhcpNetwork.__init__(self,listen_address,server_listen_port,client_listen_port)
        
        self.logger = logging.getLogger('proxydhcp')
        
        try :
            #self.dhcp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            #self.dhcp_socket.setsockopt(socket.SOL_SOCKET,socket.SO_BROADCAST,1)
            self.dhcp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            
            # --- PORT SHARING HACK ---
            # Allow binding to the same port as another process (e.g. dnsmasq)
            # dnsmasq would have to have this option set as well
            # otherwise we would need to use macvtap
            self.dhcp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            if hasattr(socket, 'SO_REUSEPORT'):
                self.dhcp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
            # ------------------------------------
            
            self.dhcp_socket.setsockopt(socket.SOL_SOCKET,socket.SO_BROADCAST,1)
            if sys.platform == 'win32':
                self.dhcp_socket.bind((self.listen_address,self.listen_port))
            else:
                ifname = interface
                if ifname:
                    SO_BINDTODEVICE = getattr(socket, 'SO_BINDTODEVICE', 25)
                    self.dhcp_socket.setsockopt(socket.SOL_SOCKET, SO_BINDTODEVICE, (ifname+'\0').encode('utf-8'))
                
                self.dhcp_socket.bind(('',self.listen_port))
        
        except socket.error as msg :
            self.logger.critical("FATAL: Error creating socket or binding to interface %s on port %s: %s" % (self.listen_address, self.listen_port, str(msg)))
            sys.exit(1)
        except Exception as e:
            self.logger.critical("FATAL: Unexpected error initializing DHCP socket: %s" % str(e))
            sys.exit(1)
        
        self.loop = True
        
    def run(self):
        while self.loop:
            try:
                self.GetNextDhcpPacket()
            except Exception:
                traceback.print_exc()
        self.log('info','Service shutdown')
    
    def log(self,level,message):
        if level == 'info':
            self.logger.info(message)
        else:
            self.logger.debug(message)
            
class DHCPD(DhcpServerBase):
    loop = True
    def __init__(self,configfile='proxy.ini',client_port=68,server_port=67):
        
        self.client_port = int(client_port)
        self.server_port = int(server_port)
        self.config = parse_config(configfile)
        DhcpServerBase.__init__(self, self.config['proxy'].get("listen_address", "0.0.0.0"), self.client_port, self.server_port, self.config['proxy'].get("interface", ""))
        self.log('info',"Starting DHCP on ports client: %s, server: %s"%(self.client_port,self.server_port))

    def HandleDhcpDiscover(self, packet):
        if packet.IsOption('vendor_class_identifier'):
            class_identifier = strlist(packet.GetOption('vendor_class_identifier'))
            if class_identifier.str()[0:9] == "PXEClient":
                responsepacket = DhcpPacket()
                responsepacket.CreateDhcpOfferPacketFrom(packet)
                responsepacket.SetMultipleOptions( {
                    'hlen': packet.GetOption("hlen"),
                    'htype': packet.GetOption("htype"),
                    'xid': packet.GetOption("xid"),
                    'flags': packet.GetOption("flags"),
                    'giaddr': packet.GetOption("giaddr"),
                    'yiaddr':[0,0,0,0],
                    'ciaddr':[0,0,0,0]
                    } )
                responsepacket.SetOption("vendor_class_identifier", b"PXEClient")
                responsepacket.SetOption("server_identifier", list(map(int, self.config['proxy']["listen_address"].split("."))))
                if self.config['proxy']['vendor_specific_information']:
                    responsepacket.SetOption('vendor_specific_information', self.config['proxy']['vendor_specific_information'].encode('ascii'))
                    
                responsepacket.DeleteOption('ip_address_lease_time')
                self.SendDhcpPacketTo(responsepacket, "255.255.255.255", self.client_port)
                self.log('info','******Responded to PXE Discover from ' + ":".join(list(map(self.fmtHex,packet.GetHardwareAddress()))))
            else:
                self.log('debug','Noticed a non-boot DHCP Discover packet from '  + ":".join(list(map(self.fmtHex,packet.GetHardwareAddress()))))
        else:
            self.log('debug','Noticed a non-boot DHCP Discover packet from '  + ":".join(list(map(self.fmtHex,packet.GetHardwareAddress()))))

    def HandleDhcpRequest(self, packet):
        self.log('debug','Noticed a DHCP Request (port 67) packet from '  + ":".join(list(map(self.fmtHex,packet.GetHardwareAddress()))))

    def HandleDhcpDecline(self, packet):
        self.log('debug','Noticed a DHCP Decline packet from '  + ":".join(list(map(self.fmtHex,packet.GetHardwareAddress()))))

    def HandleDhcpRelease(self, packet):
        self.log('debug','Noticed a DHCP Release packet from '  + ":".join(list(map(self.fmtHex,packet.GetHardwareAddress()))))

    def HandleDhcpInform(self, packet):
        self.log('debug','Noticed a DHCP Inform packet from '  + ":".join(list(map(self.fmtHex,packet.GetHardwareAddress()))))

    def fmtHex(self,input):
        return f"{input:02x}"

class ProxyDHCPD(DHCPD):
    
    def __init__(self,configfile='proxy.ini',client_port=68,server_port=4011):
        self.config = parse_config(configfile)
        self.client_port = client_port
        self.server_port = server_port
        DHCPD.__init__(self,configfile,server_port=server_port)

    def get_boot_filename(self, packet):
        default_file = self.config['proxy']['filename']
        if not packet.IsOption('client_system'):
            return default_file
            
        arch_opt = packet.GetOption('client_system')
        if not arch_opt:
            return default_file
            
        # Option 93 is unassigned natively so it comes back as a list of integers
        arch_list = arch_opt
        if hasattr(arch_opt, 'list'):
            arch_list = arch_opt.list()
            
        if len(arch_list) >= 2:
            arch = arch_list[1]
            if arch == 6:
                return self.config['proxy'].get('filename_efi32', default_file)
            elif arch in [7, 9]:
                return self.config['proxy'].get('filename_efi64', default_file)
                
        return default_file

    def HandleDhcpDiscover(self, packet):
        self.log('debug','Noticed a DHCP Discover packet from '  + ":".join(list(map(self.fmtHex,packet.GetHardwareAddress()))))
    
    def HandleDhcpRequest(self, packet):
        if packet.IsOption('vendor_class_identifier'):
            
            class_identifier = strlist(packet.GetOption('vendor_class_identifier'))
            if class_identifier.str()[0:9] == "PXEClient":
                
                # 1. CONFIGURATION PARSING WITH FALLBACKS
                ipxe_cfg = self.config.get('ipxe', {})
                ipxe_enabled = str(ipxe_cfg.get('enabled', 'False')).lower() == 'true'
                legacy_bootstrap = ipxe_cfg.get('legacy_bootstrap', 'undionly.kpxe')
                efi_bootstrap = ipxe_cfg.get('efi_bootstrap', 'ipxe.efi')
                chainload_url = ipxe_cfg.get('chainload_url', 'boot.ipxe')

                # 2. THE ROUTING LOGIC (Option 77 and Option 93)
                user_class = packet.GetOption('user_class') if packet.IsOption('user_class') else []
                client_arch = packet.GetOption('client_system') if packet.IsOption('client_system') else []

                # Convert Option 77 to a string safely
                user_class_str = ""
                if user_class:
                    user_class_list = user_class.list() if hasattr(user_class, 'list') else user_class
                    user_class_str = "".join([chr(c) for c in user_class_list if c != 0])

                # Extract Option 93 architecture list
                arch_list = []
                if client_arch:
                    arch_list = client_arch.list() if hasattr(client_arch, 'list') else client_arch

                # Determine the target boot file
                boot_file_name = None
                if ipxe_enabled:
                    if 'iPXE' in user_class_str:
                        boot_file_name = chainload_url
                        self.log('info', 'iPXE client detected. Directing to chainload URL: %s' % chainload_url)
                    elif arch_list == [0, 0]:
                        boot_file_name = legacy_bootstrap
                        self.log('info', 'Legacy BIOS hardware detected. Directing to %s' % legacy_bootstrap)
                    elif arch_list == [0, 7] or arch_list == [0, 9]:
                        boot_file_name = efi_bootstrap
                        self.log('info', 'UEFI hardware detected. Directing to %s' % efi_bootstrap)
                    else:
                        self.log('warning', 'Unsupported architecture detected: %s. Dropping payload.' % str(arch_list))
                        boot_file_name = None
                else:
                    # Fallback to standard ProxyDHCP legacy handling
                    boot_file_name = self.get_boot_filename(packet)
                
                responsepacket = DhcpPacket()
                responsepacket.CreateDhcpAckPacketFrom(packet)
                responsepacket.SetMultipleOptions( {
                    'hlen': packet.GetOption("hlen"),
                    'htype': packet.GetOption("htype"),
                    'xid': packet.GetOption("xid"),
                    'flags': packet.GetOption("flags"),
                    'giaddr': packet.GetOption("giaddr"),
                    'yiaddr':[0,0,0,0],
                    'siaddr': self.config['proxy']['tftpd'],
                    'vendor_class_identifier': b"PXEClient",
                    'server_identifier': list(map(int, self.config['proxy']["listen_address"].split("."))),
                    'tftp_server_name': self.config['proxy']['tftpd']
                } )
                
                # 3. EDGE CASE HANDLING (EMPTY OR NONE VALUES)
                if boot_file_name:
                    file_payload = boot_file_name.ljust(128, "\0").encode('ascii')
                    bootfile_name_payload = (boot_file_name + "\0").encode('ascii')
                    responsepacket.SetOption('file', file_payload)
                    responsepacket.SetOption('bootfile_name', bootfile_name_payload)
                else:
                    self.log('debug', 'boot_file_name is empty, skipping Option 67 (bootfile_name)')
                
                if self.config['proxy'].get('vendor_specific_information'):
                    responsepacket.SetOption('vendor_specific_information', self.config['proxy']['vendor_specific_information'].encode('ascii'))
                    
                responsepacket.DeleteOption('ip_address_lease_time')
                self.SendDhcpPacketTo(responsepacket, ".".join(list(map(str,packet.GetOption('ciaddr')))), self.client_port)
                self.log('info','****Responded to PXE request (port 4011) from ' + ":".join(list(map(self.fmtHex,packet.GetHardwareAddress()))))

    def HandleDhcpDecline(self, packet):
        self.log('debug','Noticed a DHCP Decline packet from '  + ":".join(list(map(self.fmtHex,packet.GetHardwareAddress()))))

    def HandleDhcpRelease(self, packet):
        self.log('debug','Noticed a DHCP Release packet from '  + ":".join(list(map(self.fmtHex,packet.GetHardwareAddress()))))

    def HandleDhcpInform(self, packet):
        self.log('debug','Noticed a DHCP Inform packet from '  + ":".join(list(map(self.fmtHex,packet.GetHardwareAddress()))))

    def fmtHex(self,input):
        return f"{input:02x}"


#!/usr/bin/python
"""
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from .dhcpd import DHCPD, ProxyDHCPD
from . import __version__
import argparse
import os
import socket
import sys
import threading
import time
import traceback

import logging
import logging.handlers


def setup_global_logger():
    logger = logging.getLogger('proxydhcp')
    logger.setLevel(logging.DEBUG)
    if not logger.handlers:
        formatter = logging.Formatter('%(asctime)s %(levelname)s ProxyDHCP: %(message)s')  
        consoleLog = logging.StreamHandler()
        consoleLog.setFormatter(formatter)
        logger.addHandler(consoleLog)
        
        if sys.platform == 'win32':
            fileLog = logging.FileHandler('proxy.log')
            fileLog.setFormatter(formatter)
            logger.addHandler(fileLog)
        else:
            if sys.platform == 'darwin':
                syslogLog = logging.handlers.SysLogHandler("/var/run/syslog")
            else:
                syslogLog = logging.handlers.SysLogHandler("/dev/log")
            syslogLog.setFormatter(formatter)
            syslogLog.setLevel(logging.INFO)
            logger.addHandler(syslogLog)


def main():
    parser = argparse.ArgumentParser(
        prog='proxydhcpd',
        description='ProxyDHCPd — A proxy DHCP server in pure Python 3 with native iPXE chainloading.',
    )
    parser.add_argument(
        '-V', '--version',
        action='version',
        version='%(prog)s ' + __version__,
    )
    parser.add_argument(
        '-c', '--config',
        default='/etc/proxydhcpd/proxy.ini',
        metavar='FILE',
        help='Path to the configuration file (default: /etc/proxydhcpd/proxy.ini)',
    )
    parser.add_argument(
        '-d', '--daemon',
        action='store_true',
        help='Run as a background daemon via double-fork (ignored on Win32)',
    )
    parser.add_argument(
        '-p', '--proxy-only',
        action='store_true',
        help='Run only the ProxyDHCP server on port 4011 (skip port 67)',
    )

    args = parser.parse_args()

    setup_global_logger()

    configfile = args.config
    daemon = args.daemon
    proxy_only = args.proxy_only

    if not os.access(configfile, os.R_OK):
        print("Unable to read config file: %s" % configfile)
        sys.exit(2)

    server = None

    # Set up a DHCPD instance
    if not proxy_only:
        try:
            server = DHCPD(configfile)
        except socket.error as msg:
            print("Error initiating on normal port, will try only 4011")
        except Exception:
            traceback.print_exc()
            print("Failed to start.")
            sys.exit(1)
        
    try:
        proxyserver = ProxyDHCPD(configfile=configfile)
    except socket.error as msg:
        print("Error initiating Proxy, already running?")
        sys.exit(1)
    except Exception:
        traceback.print_exc()
        print("Failed to start proxy.")
        sys.exit(1)
    
    # Daemonise
    if daemon:
        if sys.platform == 'win32':
            print("Ignoring request to run as daemon on Win32")
        else:
            # Do double-fork magic
            try:
                pid = os.fork()
                if pid > 0:
                    sys.exit(0)
            except OSError as e:
                print("Fork #1 failed: %d (%s)" % (e.errno, e.strerror))
                sys.exit(1)
                
            # Decouple from parent environment
            os.chdir("/")
            os.setsid()
            os.umask(0)
            
            # Do second fork
            try:
                pid = os.fork()
                if pid > 0:
                    # Exit from second parent
                    if server:
                        server.logger.info("proxy daemon has PID %d" % pid)
                        server.logger.removeHandler(server.consoleLog)
                    sys.exit(0)
            except OSError as e:
                print("Fork #2 failed: %d (%s)" % (e.errno, e.strerror))
                sys.exit(1)

    # Start loop
    if server and proxyserver:
        server.logger.info('Listening on ' + server.config['proxy']['listen_address'])
        t1 = threading.Thread(target=server.run)
        t1.daemon = True
        t1.start()
    
    t2 = threading.Thread(target=proxyserver.run)
    t2.daemon = True
    t2.start()

    while (proxyserver and proxyserver.loop) or (server and server.loop):
        try:
            time.sleep(1)
        except KeyboardInterrupt:
            print("\nExiting....")
            if proxyserver:
                proxyserver.loop = False
            if server:
                server.loop = False
            
if __name__ == "__main__":
    main()
